==> ephor_cli-0.7.1-py3-none-any.whl: ephor_cli/mcp_adapters/tools.py <==
from typing import Any, cast, get_args
import asyncio
from collections import defaultdict
import uuid
from typing import Callable, Optional

from langchain_core.tools import BaseTool, InjectedToolArg, StructuredTool, ToolException
from langchain_core.tools.base import get_all_basemodel_annotations
from mcp import ClientSession
from mcp.server.fastmcp.tools import Tool as FastMCPTool
from mcp.server.fastmcp.utilities.func_metadata import ArgModelBase, FuncMetadata
from mcp.types import (
    CallToolResult,
    EmbeddedResource,
    ImageContent,
    TextContent,
)
from mcp.types import (
    Tool as MCPTool,
)
from pydantic import BaseModel, create_model

from mcp_adapters.sessions import Connection, create_session

NonTextContent = ImageContent | EmbeddedResource
MAX_ITERATIONS = 1000

# Simplified progress tracking with direct callbacks
_progress_callbacks: dict[str, Callable] = {}

def register_progress_callback(tool_call_id: str, callback: Callable):
    """Register a callback for progress updates."""
    _progress_callbacks[tool_call_id] = callback

def unregister_progress_callback(tool_call_id: str):
    """Remove a progress callback."""
    _progress_callbacks.pop(tool_call_id, None)

def _convert_call_tool_result(
    call_tool_result: CallToolResult,
) -> tuple[str | list[str], list[NonTextContent] | None]:
    text_contents: list[TextContent] = []
    non_text_contents = []
    for content in call_tool_result.content:
        if isinstance(content, TextContent):
            text_contents.append(content)
        else:
            non_text_contents.append(content)

    tool_content: str | list[str] = [content.text for content in text_contents]
    if not text_contents:
        tool_content = ""
    elif len(text_contents) == 1:
        tool_content = tool_content[0]

    if call_tool_result.isError:
        raise ToolException(tool_content)

    return tool_content, non_text_contents or None


async def _list_all_tools(session: ClientSession) -> list[MCPTool]:
    current_cursor: str | None = None
    all_tools: list[MCPTool] = []

    iterations = 0

    while True:
        iterations += 1
        if iterations > MAX_ITERATIONS:
            raise RuntimeError("Reached max of 1000 iterations while listing tools.")

        list_tools_page_result = await session.list_tools(cursor=current_cursor)

        if list_tools_page_result.tools:
            all_tools.extend(list_tools_page_result.tools)

        if list_tools_page_result.nextCursor is None:
            break

        current_cursor = list_tools_page_result.nextCursor
    return all_tools


def convert_mcp_tool_to_langchain_tool(
    session: ClientSession | None,
    tool: MCPTool,
    *,
    connection: Connection | None = None,
    progress_callback: Optional[Callable] = None,
) -> BaseTool:
    """Convert an MCP tool to a LangChain tool.

    Args:
        session: MCP client session
        tool: MCP tool to convert
        connection: Optional connection config to use to create a new session
        progress_callback: Optional callback for progress updates

    Returns:
        a LangChain tool
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    async def call_tool(
        **arguments: dict[str, Any],
    ) -> tuple[str | list[str], list[NonTextContent] | None]:
        """Invoke the underlying MCP tool with progress tracking."""

        import logging
        logger = logging.getLogger(__name__)

        # Generate unique call ID for this tool invocation
        tool_call_id = str(uuid.uuid4())
        
        # Register progress callback if provided
        if progress_callback:
            register_progress_callback(tool_call_id, progress_callback)
        
        # Progress callback for MCP
        def _on_progress(progress: float, total: float | None, message: str | None):
            try:
                percentage = (progress / total * 100) if total else progress
                logger.info(f"[tool:{tool.name}] progress={percentage:.2f}% msg={message}")
                
                # Call the registered progress callback if available
                if tool_call_id in _progress_callbacks:
                    callback = _progress_callbacks[tool_call_id]
                    callback(tool_call_id, tool.name, percentage, message)
                    
            except Exception as err:
                logger.debug(f"Progress callback error: {err}")

        # Call the tool with progress callback
        try:
            if session is None:
                async with create_session(connection) as tool_session:
                    await tool_session.initialize()
                    call_tool_result = await cast(ClientSession, tool_session).call_tool(
                        tool.name,
                        arguments,
                        progress_callback=_on_progress,
                    )
            else:
                call_tool_result = await session.call_tool(
                    tool.name,
                    arguments,
                    progress_callback=_on_progress,
                )

            # Convert and return the result
            return _convert_call_tool_result(call_tool_result)
        finally:
            # Clean up progress callback
            unregister_progress_callback(tool_call_id)

    return StructuredTool(
        name=tool.name,
        description=tool.description or "",
        args_schema=tool.inputSchema,
        coroutine=call_tool,
        response_format="content_and_artifact",
        metadata=tool.annotations.model_dump() if tool.annotations else None,
    )


async def load_mcp_tools(
    session: ClientSession | None,
    *,
    connection: Connection | None = None,
    progress_callback: Optional[Callable] = None,
) -> list[BaseTool]:
    """Load all available MCP tools and convert them to LangChain tools.

    Args:
        session: MCP client session
        connection: Optional connection config
        progress_callback: Optional callback for progress updates

    Returns:
        list of LangChain tools with progress tracking
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    if session is None:
        # If a session is not provided, we will create one on the fly
        async with create_session(connection) as tool_session:
            await tool_session.initialize()
            tools = await _list_all_tools(tool_session)
    else:
        tools = await _list_all_tools(session)

    converted_tools = [
        convert_mcp_tool_to_langchain_tool(
            session, 
            tool, 
            connection=connection,
            progress_callback=progress_callback
        ) for tool in tools
    ]
    return converted_tools


def _get_injected_args(tool: BaseTool) -> list[str]:
    def _is_injected_arg_type(type_: type) -> bool:
        return any(
            isinstance(arg, InjectedToolArg)
            or (isinstance(arg, type) and issubclass(arg, InjectedToolArg))
            for arg in get_args(type_)[1:]
        )

    injected_args = [
        field
        for field, field_info in get_all_basemodel_annotations(tool.args_schema).items()
        if _is_injected_arg_type(field_info)
    ]
    return injected_args


def to_fastmcp(tool: BaseTool) -> FastMCPTool:
    """Convert a LangChain tool to a FastMCP tool."""
    if not issubclass(tool.args_schema, BaseModel):
        raise ValueError(
            "Tool args_schema must be a subclass of pydantic.BaseModel. "
            "Tools with dict args schema are not supported."
        )

    parameters = tool.tool_call_schema.model_json_schema()
    field_definitions = {
        field: (field_info.annotation, field_info)
        for field, field_info in tool.tool_call_schema.model_fields.items()
    }
    arg_model = create_model(
        f"{tool.name}Arguments",
        **field_definitions,
        __base__=ArgModelBase,
    )
    fn_metadata = FuncMetadata(arg_model=arg_model)

    async def fn(**arguments: dict[str, Any]) -> Any:
        return await tool.ainvoke(arguments)

    injected_args = _get_injected_args(tool)
    if len(injected_args) > 0:
        raise NotImplementedError("LangChain tools with injected arguments are not supported")

    fastmcp_tool = FastMCPTool(
        fn=fn,
        name=tool.name,
        description=tool.description,
        parameters=parameters,
        fn_metadata=fn_metadata,
        is_async=True,
    )
    return fastmcp_tool

__all__ = [
    "load_mcp_tools",
    "convert_mcp_tool_to_langchain_tool", 
    "to_fastmcp",
    "register_progress_callback",
    "unregister_progress_callback",
]

==> ephor_cli-0.7.3-py3-none-any.whl: ephor_cli/mcp_adapters/tools.py <==
from typing import Any, cast, get_args
import asyncio
from collections import defaultdict
import uuid
from typing import Callable, Optional

from langchain_core.tools import BaseTool, InjectedToolArg, StructuredTool, ToolException
from langchain_core.tools.base import get_all_basemodel_annotations
from mcp import ClientSession
from mcp.server.fastmcp.tools import Tool as FastMCPTool
from mcp.server.fastmcp.utilities.func_metadata import ArgModelBase, FuncMetadata
from mcp.types import (
    CallToolResult,
    EmbeddedResource,
    ImageContent,
    TextContent,
)
from mcp.types import (
    Tool as MCPTool,
)
from pydantic import BaseModel, create_model

from .sessions import Connection, create_session

NonTextContent = ImageContent | EmbeddedResource
MAX_ITERATIONS = 1000

# Simplified progress tracking with direct callbacks
_progress_callbacks: dict[str, Callable] = {}

def register_progress_callback(tool_call_id: str, callback: Callable):
    """Register a callback for progress updates."""
    _progress_callbacks[tool_call_id] = callback

def unregister_progress_callback(tool_call_id: str):
    """Remove a progress callback."""
    _progress_callbacks.pop(tool_call_id, None)

def _convert_call_tool_result(
    call_tool_result: CallToolResult,
) -> tuple[str | list[str], list[NonTextContent] | None]:
    text_contents: list[TextContent] = []
    non_text_contents = []
    for content in call_tool_result.content:
        if isinstance(content, TextContent):
            text_contents.append(content)
        else:
            non_text_contents.append(content)

    tool_content: str | list[str] = [content.text for content in text_contents]
    if not text_contents:
        tool_content = ""
    elif len(text_contents) == 1:
        tool_content = tool_content[0]

    if call_tool_result.isError:
        raise ToolException(tool_content)

    return tool_content, non_text_contents or None


async def _list_all_tools(session: ClientSession) -> list[MCPTool]:
    current_cursor: str | None = None
    all_tools: list[MCPTool] = []

    iterations = 0

    while True:
        iterations += 1
        if iterations > MAX_ITERATIONS:
            raise RuntimeError("Reached max of 1000 iterations while listing tools.")

        list_tools_page_result = await session.list_tools(cursor=current_cursor)

        if list_tools_page_result.tools:
            all_tools.extend(list_tools_page_result.tools)

        if list_tools_page_result.nextCursor is None:
            break

        current_cursor = list_tools_page_result.nextCursor
    return all_tools


def convert_mcp_tool_to_langchain_tool(
    session: ClientSession | None,
    tool: MCPTool,
    *,
    connection: Connection | None = None,
    progress_callback: Optional[Callable] = None,
) -> BaseTool:
    """Convert an MCP tool to a LangChain tool.

    Args:
        session: MCP client session
        tool: MCP tool to convert
        connection: Optional connection config to use to create a new session
        progress_callback: Optional callback for progress updates

    Returns:
        a LangChain tool
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    async def call_tool(
        **arguments: dict[str, Any],
    ) -> tuple[str | list[str], list[NonTextContent] | None]:
        """Invoke the underlying MCP tool with progress tracking."""

        import logging
        logger = logging.getLogger(__name__)

        # Generate unique call ID for this tool invocation
        tool_call_id = str(uuid.uuid4())
        
        # Register progress callback if provided
        if progress_callback:
            register_progress_callback(tool_call_id, progress_callback)
        
        # Progress callback for MCP
        def _on_progress(progress: float, total: float | None, message: str | None):
            try:
                percentage = (progress / total * 100) if total else progress
                logger.info(f"[tool:{tool.name}] progress={percentage:.2f}% msg={message}")
                
                # Call the registered progress callback if available
                if tool_call_id in _progress_callbacks:
                    callback = _progress_callbacks[tool_call_id]
                    callback(tool_call_id, tool.name, percentage, message)
                    
            except Exception as err:
                logger.debug(f"Progress callback error: {err}")

        # Call the tool with progress callback
        try:
            if session is None:
                async with create_session(connection) as tool_session:
                    await tool_session.initialize()
                    call_tool_result = await cast(ClientSession, tool_session).call_tool(
                        tool.name,
                        arguments,
                        progress_callback=_on_progress,
                    )
            else:
                call_tool_result = await session.call_tool(
                    tool.name,
                    arguments,
                    progress_callback=_on_progress,
                )

            # Convert and return the result
            return _convert_call_tool_result(call_tool_result)
        finally:
            # Clean up progress callback
            unregister_progress_callback(tool_call_id)

    return StructuredTool(
        name=tool.name,
        description=tool.description or "",
        args_schema=tool.inputSchema,
        coroutine=call_tool,
        response_format="content_and_artifact",
        metadata=tool.annotations.model_dump() if tool.annotations else None,
    )


async def load_mcp_tools(
    session: ClientSession | None,
    *,
    connection: Connection | None = None,
    progress_callback: Optional[Callable] = None,
) -> list[BaseTool]:
    """Load all available MCP tools and convert them to LangChain tools.

    Args:
        session: MCP client session
        connection: Optional connection config
        progress_callback: Optional callback for progress updates

    Returns:
        list of LangChain tools with progress tracking
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    if session is None:
        # If a session is not provided, we will create one on the fly
        async with create_session(connection) as tool_session:
            await tool_session.initialize()
            tools = await _list_all_tools(tool_session)
    else:
        tools = await _list_all_tools(session)

    converted_tools = [
        convert_mcp_tool_to_langchain_tool(
            session, 
            tool, 
            connection=connection,
            progress_callback=progress_callback
        ) for tool in tools
    ]
    return converted_tools


def _get_injected_args(tool: BaseTool) -> list[str]:
    def _is_injected_arg_type(type_: type) -> bool:
        return any(
            isinstance(arg, InjectedToolArg)
            or (isinstance(arg, type) and issubclass(arg, InjectedToolArg))
            for arg in get_args(type_)[1:]
        )

    injected_args = [
        field
        for field, field_info in get_all_basemodel_annotations(tool.args_schema).items()
        if _is_injected_arg_type(field_info)
    ]
    return injected_args


def to_fastmcp(tool: BaseTool) -> FastMCPTool:
    """Convert a LangChain tool to a FastMCP tool."""
    if not issubclass(tool.args_schema, BaseModel):
        raise ValueError(
            "Tool args_schema must be a subclass of pydantic.BaseModel. "
            "Tools with dict args schema are not supported."
        )

    parameters = tool.tool_call_schema.model_json_schema()
    field_definitions = {
        field: (field_info.annotation, field_info)
        for field, field_info in tool.tool_call_schema.model_fields.items()
    }
    arg_model = create_model(
        f"{tool.name}Arguments",
        **field_definitions,
        __base__=ArgModelBase,
    )
    fn_metadata = FuncMetadata(arg_model=arg_model)

    async def fn(**arguments: dict[str, Any]) -> Any:
        return await tool.ainvoke(arguments)

    injected_args = _get_injected_args(tool)
    if len(injected_args) > 0:
        raise NotImplementedError("LangChain tools with injected arguments are not supported")

    fastmcp_tool = FastMCPTool(
        fn=fn,
        name=tool.name,
        description=tool.description,
        parameters=parameters,
        fn_metadata=fn_metadata,
        is_async=True,
    )
    return fastmcp_tool

__all__ = [
    "load_mcp_tools",
    "convert_mcp_tool_to_langchain_tool", 
    "to_fastmcp",
    "register_progress_callback",
    "unregister_progress_callback",
]
